@@ -29,9 +29,12 @@ typedef struct TestRequestParam
     const char *securityToken;
     const char *range;
     const char *kms;
+    const char *sseC;
     const char *ttl;
     const char *token;
     const char *tag;
+    bool requesterPays;
+    bool contentMd5;
 } TestRequestParam;
 
 #define testRequestP(write, s3, verb, path, ...)                                                                                   \
@@ -66,7 +69,7 @@ testRequest(IoWrite *write, Storage *s3, const char *verb, const char *path, Tes
             "authorization:AWS4-HMAC-SHA256 Credential=%s/\?\?\?\?\?\?\?\?/us-east-1/s3/aws4_request,SignedHeaders=",
             param.accessKey == NULL ? strZ(driver->accessKey) : param.accessKey);
 
-        if (param.content != NULL)
+        if (param.contentMd5)
             strCatZ(request, "content-md5;");
 
         strCatZ(request, "host;");
@@ -76,12 +79,23 @@ testRequest(IoWrite *write, Storage *s3, const char *verb, const char *path, Tes
 
         strCatZ(request, "x-amz-content-sha256;x-amz-date");
 
+        if (param.requesterPays)
+            strCatZ(request, ";x-amz-request-payer");
+
         if (securityToken != NULL)
             strCatZ(request, ";x-amz-security-token");
 
         if (param.kms != NULL)
             strCatZ(request, ";x-amz-server-side-encryption;x-amz-server-side-encryption-aws-kms-key-id");
 
+        if (param.sseC != NULL)
+        {
+            strCatZ(
+                request,
+                ";x-amz-server-side-encryption-customer-algorithm;x-amz-server-side-encryption-customer-key"
+                ";x-amz-server-side-encryption-customer-key-md5");
+        }
+
         if (param.tag != NULL)
             strCatZ(request, ";x-amz-tagging");
 
@@ -92,7 +106,7 @@ testRequest(IoWrite *write, Storage *s3, const char *verb, const char *path, Tes
     strCatFmt(request, "content-length:%zu\r\n", param.content != NULL ? strlen(param.content) : 0);
 
     // Add md5
-    if (param.content != NULL)
+    if (param.contentMd5)
     {
         strCatFmt(
             request, "content-md5:%s\r\n", strZ(strNewEncode(encodingBase64, cryptoHashOne(hashTypeMd5, BUFSTRZ(param.content)))));
@@ -129,6 +143,10 @@ testRequest(IoWrite *write, Storage *s3, const char *verb, const char *path, Tes
             strCatFmt(request, "x-amz-security-token:%s\r\n", securityToken);
     }
 
+    // Requestor pays
+    if (param.requesterPays)
+        strCatZ(request, "x-amz-request-payer:requester\r\n");
+
     // Add kms key
     if (param.kms != NULL)
     {
@@ -136,6 +154,16 @@ testRequest(IoWrite *write, Storage *s3, const char *verb, const char *path, Tes
         strCatFmt(request, "x-amz-server-side-encryption-aws-kms-key-id:%s\r\n", param.kms);
     }
 
+    // Add sseC key
+    if (param.sseC != NULL)
+    {
+        strCatZ(request, "x-amz-server-side-encryption-customer-algorithm:AES256\r\n");
+        strCatFmt(request, "x-amz-server-side-encryption-customer-key:%s\r\n", param.sseC);
+        strCatFmt(
+            request, "x-amz-server-side-encryption-customer-key-md5:%s\r\n",
+            strZ(strNewEncode(encodingBase64, cryptoHashOne(hashTypeMd5, bufNewDecode(encodingBase64, STR(param.sseC))))));
+    }
+
     // Add tags
     if (param.tag != NULL)
         strCatFmt(request, "x-amz-tagging:%s\r\n", param.tag);
@@ -223,13 +251,7 @@ testS3DateTime(time_t time)
         FUNCTION_HARNESS_PARAM(TIME, time);
     FUNCTION_HARNESS_END();
 
-    char buffer[21];
-
-    THROW_ON_SYS_ERROR(
-        strftime(buffer, sizeof(buffer), "%Y-%m-%dT%H:%M:%SZ", gmtime(&time)) != sizeof(buffer) - 1, AssertError,
-        "unable to format date");
-
-    FUNCTION_HARNESS_RETURN(STRING, strNewZ(buffer));
+    FUNCTION_HARNESS_RETURN(STRING, strNewTimeP("%Y-%m-%dT%H:%M:%SZ", time, .utc = true));
 }
 
 /***********************************************************************************************************************************
@@ -250,8 +272,6 @@ testRun(void)
     const String *region = STRDEF("us-east-1");
     const String *endPoint = STRDEF("s3.amazonaws.com");
     const String *host = hrnServerHost();
-    const unsigned int port = hrnServerPort(0);
-    const unsigned int authPort = hrnServerPort(1);
     const String *accessKey = STRDEF("AKIAIOSFODNN7EXAMPLE");
     const String *secretAccessKey = STRDEF("wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY");
     const String *securityToken = STRDEF(
@@ -281,15 +301,12 @@ testRun(void)
     hrnCfgEnvRaw(cfgOptRepoS3KeySecret, secretAccessKey);
 
     // *****************************************************************************************************************************
-    if (testBegin("storageS3DateTime() and storageS3Auth()"))
+    if (testBegin("storageS3Auth()"))
     {
-        char logBuf[STACK_TRACE_PARAM_MAX];
-
-        TEST_RESULT_STR_Z(storageS3DateTime(1491267845), "20170404T010405Z", "static date");
-
         // -------------------------------------------------------------------------------------------------------------------------
         TEST_TITLE("config without token");
 
+        char logBuf[STACK_TRACE_PARAM_MAX];
         StringList *argList = strLstDup(commonArgList);
         HRN_CFG_LOAD(cfgCmdArchivePush, argList);
 
@@ -404,16 +421,18 @@ testRun(void)
     {
         HRN_FORK_BEGIN()
         {
+            const unsigned int testPort = hrnServerPortNext();
+            const unsigned int testPortAuth = hrnServerPortNext();
+
             HRN_FORK_CHILD_BEGIN(.prefix = "s3 server", .timeout = 5000)
             {
-                TEST_RESULT_VOID(hrnServerRunP(HRN_FORK_CHILD_READ(), hrnServerProtocolTls, .port = port), "s3 server run");
+                TEST_RESULT_VOID(hrnServerRunP(HRN_FORK_CHILD_READ(), hrnServerProtocolTls, testPort), "s3 server");
             }
             HRN_FORK_CHILD_END();
 
             HRN_FORK_CHILD_BEGIN(.prefix = "auth server", .timeout = 10000)
             {
-                TEST_RESULT_VOID(
-                    hrnServerRunP(HRN_FORK_CHILD_READ(), hrnServerProtocolSocket, .port = authPort), "auth server run");
+                TEST_RESULT_VOID(hrnServerRunP(HRN_FORK_CHILD_READ(), hrnServerProtocolSocket, testPortAuth), "auth server");
             }
             HRN_FORK_CHILD_END();
 
@@ -429,7 +448,7 @@ testRun(void)
                 TEST_TITLE("config with keys, token, and host with custom port");
 
                 StringList *argList = strLstDup(commonArgList);
-                hrnCfgArgRawFmt(argList, cfgOptRepoStorageHost, "%s:%u", strZ(host), port);
+                hrnCfgArgRawFmt(argList, cfgOptRepoStorageHost, "%s:%u", strZ(host), testPort);
                 hrnCfgEnvRaw(cfgOptRepoS3Token, securityToken);
                 HRN_CFG_LOAD(cfgCmdArchivePush, argList);
 
@@ -488,13 +507,15 @@ testRun(void)
                 hrnServerScriptClose(service);
 
                 argList = strLstDup(commonArgList);
-                hrnCfgArgRawFmt(argList, cfgOptRepoStorageHost, "%s:%u", strZ(host), port);
+                hrnCfgArgRawFmt(argList, cfgOptRepoStorageHost, "%s:%u", strZ(host), testPort);
                 hrnCfgArgRaw(argList, cfgOptRepoS3Role, credRole);
                 hrnCfgArgRawStrId(argList, cfgOptRepoS3KeyType, storageS3KeyTypeAuto);
                 hrnCfgArgRawZ(argList, cfgOptRepoS3KmsKeyId, "kmskey1");
+                hrnCfgEnvRawZ(cfgOptRepoS3SseCustomerKey, "rA1P");
                 hrnCfgArgRawZ(argList, cfgOptRepoStorageTag, "Key1=Value1");
                 hrnCfgArgRawZ(argList, cfgOptRepoStorageTag, " Key 2= Value 2");
                 HRN_CFG_LOAD(cfgCmdArchivePush, argList);
+                hrnCfgEnvRemoveRaw(cfgOptRepoS3SseCustomerKey);
 
                 s3 = storageRepoGet(0, true);
                 driver = (StorageS3 *)storageDriver(s3);
@@ -508,7 +529,7 @@ testRun(void)
 
                 // Testing requires the auth http client to be redirected
                 driver->credHost = hrnServerHost();
-                driver->credHttpClient = httpClientNew(sckClientNew(host, authPort, 5000, 5000), 5000);
+                driver->credHttpClient = httpClientNew(sckClientNew(host, testPortAuth, 5000, 5000), 5000);
 
                 // Now that we have checked the role when set explicitly, null it out to make sure it is retrieved automatically
                 driver->credRole = NULL;
@@ -660,7 +681,7 @@ testRun(void)
 
                 hrnServerScriptClose(auth);
 
-                testRequestP(service, s3, HTTP_VERB_GET, "/file.txt", .accessKey = "x", .securityToken = "z");
+                testRequestP(service, s3, HTTP_VERB_GET, "/file.txt", .accessKey = "x", .securityToken = "z", .sseC = "rA1P");
                 testResponseP(service, .code = 303, .content = "CONTENT");
 
                 StorageRead *read = NULL;
@@ -680,6 +701,9 @@ testRun(void)
                     "x-amz-content-sha256: e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855\n"
                     "x-amz-date: <redacted>\n"
                     "x-amz-security-token: <redacted>\n"
+                    "x-amz-server-side-encryption-customer-algorithm: AES256\n"
+                    "x-amz-server-side-encryption-customer-key: <redacted>\n"
+                    "x-amz-server-side-encryption-customer-key-md5: <redacted>\n"
                     "*** Response Headers ***:\n"
                     "content-length: 7\n"
                     "*** Response Content ***:\n"
@@ -713,7 +737,7 @@ testRun(void)
 
                 testRequestP(
                     service, s3, HTTP_VERB_PUT, "/file.txt", .content = "ABCD", .accessKey = "xx", .securityToken = "zz",
-                    .kms = "kmskey1", .tag = "%20Key%202=%20Value%202&Key1=Value1");
+                    .kms = "kmskey1", .sseC = "rA1P", .tag = "%20Key%202=%20Value%202&Key1=Value1");
                 testResponseP(service);
 
                 // Make a copy of the signing key to verify that it gets changed when the keys are updated
@@ -732,7 +756,7 @@ testRun(void)
                 TEST_RESULT_BOOL(storageWriteSyncPath(write), true, "path is synced");
                 TEST_RESULT_BOOL(storageWriteTruncate(write), true, "file will be truncated");
 
-                TEST_RESULT_VOID(storageWriteS3Close(write->driver), "close file again");
+                TEST_RESULT_VOID(storageWriteS3Close(ioWriteDriver(storageWriteIo(write))), "close file again");
 
                 // Check that temp credentials were changed
                 TEST_RESULT_STR_Z(driver->accessKey, "xx", "check access key");
@@ -746,7 +770,7 @@ testRun(void)
                 TEST_TITLE("write zero-length file");
 
                 testRequestP(
-                    service, s3, HTTP_VERB_PUT, "/file.txt", .content = "", .kms = "kmskey1",
+                    service, s3, HTTP_VERB_PUT, "/file.txt", .content = "", .kms = "kmskey1", .sseC = "rA1P",
                     .tag = "%20Key%202=%20Value%202&Key1=Value1");
                 testResponseP(service);
 
@@ -757,7 +781,7 @@ testRun(void)
                 TEST_TITLE("write file in chunks with nothing left over on close");
 
                 testRequestP(
-                    service, s3, HTTP_VERB_POST, "/file.txt?uploads=", .kms = "kmskey1",
+                    service, s3, HTTP_VERB_POST, "/file.txt?uploads=", .kms = "kmskey1", .sseC = "rA1P",
                     .tag = "%20Key%202=%20Value%202&Key1=Value1");
                 testResponseP(
                     service,
@@ -769,10 +793,14 @@ testRun(void)
                         "<UploadId>WxRt</UploadId>"
                         "</InitiateMultipartUploadResult>");
 
-                testRequestP(service, s3, HTTP_VERB_PUT, "/file.txt?partNumber=1&uploadId=WxRt", .content = "1234567890123456");
+                testRequestP(
+                    service, s3, HTTP_VERB_PUT, "/file.txt?partNumber=1&uploadId=WxRt", .content = "1234567890123456",
+                    .sseC = "rA1P");
                 testResponseP(service, .header = "etag:WxRt1");
 
-                testRequestP(service, s3, HTTP_VERB_PUT, "/file.txt?partNumber=2&uploadId=WxRt", .content = "7890123456789012");
+                testRequestP(
+                    service, s3, HTTP_VERB_PUT, "/file.txt?partNumber=2&uploadId=WxRt", .content = "7890123456789012",
+                    .sseC = "rA1P");
                 testResponseP(service, .header = "eTag:WxRt2");
 
                 testRequestP(
@@ -798,7 +826,7 @@ testRun(void)
                 // Stop writing tags
                 driver->tag = NULL;
 
-                testRequestP(service, s3, HTTP_VERB_POST, "/file.txt?uploads=", .kms = "kmskey1");
+                testRequestP(service, s3, HTTP_VERB_POST, "/file.txt?uploads=", .kms = "kmskey1", .sseC = "rA1P");
                 testResponseP(
                     service,
                     .content =
@@ -809,10 +837,14 @@ testRun(void)
                         "<UploadId>WxRt</UploadId>"
                         "</InitiateMultipartUploadResult>");
 
-                testRequestP(service, s3, HTTP_VERB_PUT, "/file.txt?partNumber=1&uploadId=WxRt", .content = "1234567890123456");
+                testRequestP(
+                    service, s3, HTTP_VERB_PUT, "/file.txt?partNumber=1&uploadId=WxRt", .content = "1234567890123456",
+                    .sseC = "rA1P");
                 testResponseP(service, .header = "etag:WxRt1");
 
-                testRequestP(service, s3, HTTP_VERB_PUT, "/file.txt?partNumber=2&uploadId=WxRt", .content = "7890123456789012");
+                testRequestP(
+                    service, s3, HTTP_VERB_PUT, "/file.txt?partNumber=2&uploadId=WxRt", .content = "7890123456789012",
+                    .sseC = "rA1P");
                 testResponseP(service, .header = "eTag:WxRt2");
 
                 testRequestP(
@@ -838,7 +870,6 @@ testRun(void)
                     "*** Request Headers ***:\n"
                     "authorization: <redacted>\n"
                     "content-length: 205\n"
-                    "content-md5: 37smUM6Ah2/EjZbp420dPw==\n"
                     "host: bucket.s3.amazonaws.com\n"
                     "x-amz-content-sha256: 0838a79dfbddc2128d28fb4fa8d605e0a8e6d1355094000f39b6eb3feff4641f\n"
                     "x-amz-date: <redacted>\n"
@@ -852,7 +883,7 @@ testRun(void)
                 // -----------------------------------------------------------------------------------------------------------------
                 TEST_TITLE("write file in chunks with something left over on close");
 
-                testRequestP(service, s3, HTTP_VERB_POST, "/file.txt?uploads=", .kms = "kmskey1");
+                testRequestP(service, s3, HTTP_VERB_POST, "/file.txt?uploads=", .kms = "kmskey1", .sseC = "rA1P");
                 testResponseP(
                     service,
                     .content =
@@ -863,10 +894,14 @@ testRun(void)
                         "<UploadId>RR55</UploadId>"
                         "</InitiateMultipartUploadResult>");
 
-                testRequestP(service, s3, HTTP_VERB_PUT, "/file.txt?partNumber=1&uploadId=RR55", .content = "1234567890123456");
+                testRequestP(
+                    service, s3, HTTP_VERB_PUT, "/file.txt?partNumber=1&uploadId=RR55", .content = "1234567890123456",
+                    .sseC = "rA1P");
                 testResponseP(service, .header = "etag:RR551");
 
-                testRequestP(service, s3, HTTP_VERB_PUT, "/file.txt?partNumber=2&uploadId=RR55", .content = "7890");
+                testRequestP(
+                    service, s3, HTTP_VERB_PUT, "/file.txt?partNumber=2&uploadId=RR55", .content = "7890",
+                    .sseC = "rA1P");
                 testResponseP(service, .header = "eTag:RR552");
 
                 testRequestP(
@@ -889,7 +924,7 @@ testRun(void)
                 // -----------------------------------------------------------------------------------------------------------------
                 TEST_TITLE("file missing");
 
-                testRequestP(service, s3, HTTP_VERB_HEAD, "/BOGUS");
+                testRequestP(service, s3, HTTP_VERB_HEAD, "/BOGUS", .sseC = "rA1P");
                 testResponseP(service, .code = 404);
 
                 TEST_RESULT_BOOL(storageExistsP(s3, STRDEF("BOGUS")), false, "check");
@@ -906,6 +941,7 @@ testRun(void)
 
                 #define TEST_SERVICE_ROLE                           "arn:aws:iam::123456789012:role/TestRole"
                 #define TEST_SERVICE_TOKEN                          "TOKEN"
+                #define TEST_SERVICE_TOKEN_FILE                     TEST_PATH "/web-id-token"
                 #define TEST_SERVICE_URI                                                                                           \
                     "/?Action=AssumeRoleWithWebIdentity&RoleArn=arn%3Aaws%3Aiam%3A%3A123456789012%3Arole%2FTestRole"               \
                         "&RoleSessionName=pgBackRest&Version=2011-06-15&WebIdentityToken=" TEST_SERVICE_TOKEN
@@ -923,10 +959,10 @@ testRun(void)
                     "</AssumeRoleWithWebIdentityResponse>"
                 // {uncrustify_on}
 
-                HRN_STORAGE_PUT_Z(storagePosixNewP(TEST_PATH_STR, .write = true), "web-id-token", TEST_SERVICE_TOKEN);
+                HRN_STORAGE_PUT_Z(storagePosixNewP(TEST_PATH_STR, .write = true), TEST_SERVICE_TOKEN_FILE, TEST_SERVICE_TOKEN);
 
                 argList = strLstDup(commonArgList);
-                hrnCfgArgRawFmt(argList, cfgOptRepoStorageHost, "%s:%u", strZ(host), port);
+                hrnCfgArgRawFmt(argList, cfgOptRepoStorageHost, "%s:%u", strZ(host), testPort);
                 hrnCfgArgRawStrId(argList, cfgOptRepoS3KeyType, storageS3KeyTypeWebId);
                 HRN_CFG_LOAD(cfgCmdArchivePush, argList);
 
@@ -940,20 +976,20 @@ testRun(void)
                     storageRepoGet(0, true), OptionInvalidError,
                     "option 'repo1-s3-key-type' is 'web-id' but 'AWS_ROLE_ARN' and 'AWS_WEB_IDENTITY_TOKEN_FILE' are not set");
 
-                setenv("AWS_WEB_IDENTITY_TOKEN_FILE", TEST_PATH "/web-id-token", true);
+                setenv("AWS_WEB_IDENTITY_TOKEN_FILE", TEST_SERVICE_TOKEN_FILE, true);
 
                 s3 = storageRepoGet(0, true);
                 driver = (StorageS3 *)storageDriver(s3);
 
                 TEST_RESULT_STR_Z(driver->credRole, TEST_SERVICE_ROLE, "check role");
-                TEST_RESULT_STR_Z(driver->webIdToken, TEST_SERVICE_TOKEN, "check token");
+                TEST_RESULT_STR_Z(driver->webIdTokenFile, TEST_SERVICE_TOKEN_FILE, "check token file");
 
                 // Set partSize to a small value for testing
                 driver->partSize = 16;
 
                 // Testing requires the auth http client to be redirected
                 driver->credHost = hrnServerHost();
-                driver->credHttpClient = httpClientNew(sckClientNew(host, authPort, 5000, 5000), 5000);
+                driver->credHttpClient = httpClientNew(sckClientNew(host, testPortAuth, 5000, 5000), 5000);
 
                 hrnServerScriptAccept(service);
 
@@ -1243,7 +1279,7 @@ testRun(void)
                 argList = strLstDup(commonArgList);
                 hrnCfgArgRawStrId(argList, cfgOptRepoS3UriStyle, storageS3UriStylePath);
                 hrnCfgArgRawFmt(argList, cfgOptRepoStorageHost, "https://%s", strZ(host));
-                hrnCfgArgRawFmt(argList, cfgOptRepoStoragePort, "%u", port);
+                hrnCfgArgRawFmt(argList, cfgOptRepoStoragePort, "%u", testPort);
                 hrnCfgEnvRemoveRaw(cfgOptRepoS3Token);
                 HRN_CFG_LOAD(cfgCmdArchivePush, argList);
 
@@ -1281,7 +1317,7 @@ testRun(void)
                         "</ListBucketResult>");
 
                 testRequestP(
-                    service, s3, HTTP_VERB_POST, "/bucket/?delete=",
+                    service, s3, HTTP_VERB_POST, "/bucket/?delete=", .contentMd5 = true,
                     .content =
                         "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                         "<Delete><Quiet>true</Quiet>"
@@ -1342,7 +1378,7 @@ testRun(void)
                         "</ListBucketResult>");
 
                 testRequestP(
-                    service, s3, HTTP_VERB_POST, "/bucket/?delete=",
+                    service, s3, HTTP_VERB_POST, "/bucket/?delete=", .contentMd5 = true,
                     .content =
                         "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                         "<Delete><Quiet>true</Quiet>"
@@ -1352,7 +1388,7 @@ testRun(void)
                 testResponseP(service);
 
                 testRequestP(
-                    service, s3, HTTP_VERB_POST, "/bucket/?delete=",
+                    service, s3, HTTP_VERB_POST, "/bucket/?delete=", .contentMd5 = true,
                     .content =
                         "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                         "<Delete><Quiet>true</Quiet>"
@@ -1381,7 +1417,7 @@ testRun(void)
                         "</ListBucketResult>");
 
                 testRequestP(
-                    service, s3, HTTP_VERB_POST, "/bucket/?delete=",
+                    service, s3, HTTP_VERB_POST, "/bucket/?delete=", .contentMd5 = true,
                     .content =
                         "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                         "<Delete><Quiet>true</Quiet>"
@@ -1409,6 +1445,135 @@ testRun(void)
 
                 TEST_RESULT_VOID(storageRemoveP(s3, STRDEF("/path/to/test.txt")), "remove");
 
+                // -----------------------------------------------------------------------------------------------------------------
+                TEST_TITLE("switch to time limited");
+
+                hrnServerScriptClose(service);
+
+                argList = strLstDup(commonArgList);
+                hrnCfgArgRawFmt(argList, cfgOptRepoStorageHost, "%s:%u", strZ(host), testPort);
+                hrnCfgArgRawZ(argList, cfgOptPgPath, "/pg1");
+                hrnCfgArgRawZ(argList, cfgOptRepo, "1");
+                hrnCfgArgRawZ(argList, cfgOptRepoTargetTime, "2024-08-04 02:54:09+00");
+                hrnCfgArgRawBool(argList, cfgOptRepoS3RequesterPays, true);
+                HRN_CFG_LOAD(cfgCmdArchiveGet, argList);
+
+                s3 = storageRepoGet(0, false);
+
+                hrnServerScriptAccept(service);
+
+                // -----------------------------------------------------------------------------------------------------------------
+                TEST_TITLE("list with time limit");
+
+                testRequestP(service, s3, HTTP_VERB_GET, "/?delimiter=%2F&prefix=path%2Fto%2F&versions=", .requesterPays = true);
+                testResponseP(
+                    service,
+                    .content =
+                        "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
+                        "<ListBucketResult xmlns=\"http://s3.amazonaws.com/doc/2006-03-01/\">"
+                        "    <IsTruncated>true</IsTruncated>"
+                        "    <NextContinuationToken>1ueG</NextContinuationToken>"
+                        "    <DeleteMarker>"
+                        "        <Key>path/to/test_file</Key>"
+                        "        <LastModified>2024-08-04T02:54:10.000Z</LastModified>"
+                        "    </DeleteMarker>"
+                        "    <Version>"
+                        "        <Key>path/to/test_file</Key>"
+                        "        <LastModified>2024-08-04T02:54:10.000Z</LastModified>"
+                        "    </Version>"
+                        "</ListBucketResult>");
+
+                testRequestP(
+                    service, s3, HTTP_VERB_GET, "/?continuation-token=1ueG&delimiter=%2F&prefix=path%2Fto%2F&versions=",
+                    .requesterPays = true);
+                testResponseP(
+                    service,
+                    .content =
+                        "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
+                        "<ListBucketResult xmlns=\"http://s3.amazonaws.com/doc/2006-03-01/\">"
+                        "    <IsTruncated>false</IsTruncated>"
+                        "    <Version>"
+                        "        <Key>path/to/test_file</Key>"
+                        "        <LastModified>2024-08-04T02:54:09.000Z</LastModified>"
+                        "        <Size>737</Size>"
+                        "        <VersionId>bbbb</VersionId>"
+                        "    </Version>"
+                        "    <Version>"
+                        "        <Key>path/to/test_file</Key>"
+                        "        <LastModified>2024-08-04T02:54:09.000Z</LastModified>"
+                        "    </Version>"
+                        "    <Version>"
+                        "        <Key>path/to/test_file2</Key>"
+                        "        <LastModified>2024-08-04T02:54:10.000Z</LastModified>"
+                        "    </Version>"
+                        "   <CommonPrefixes>"
+                        "       <Prefix>path/to/test_path/</Prefix>"
+                        "   </CommonPrefixes>"
+                        "</ListBucketResult>");
+
+                TEST_STORAGE_LIST(
+                    s3, "/path/to",
+                    "test_file {s=737, t=1722740049, v=bbbb}\n"
+                    "test_path/\n",
+                    .level = storageInfoLevelBasic, .noRecurse = true);
+
+                // -----------------------------------------------------------------------------------------------------------------
+                TEST_TITLE("get file with time limit");
+
+                testRequestP(service, s3, HTTP_VERB_GET, "/?delimiter=%2F&prefix=path%2F3%2F&versions=", .requesterPays = true);
+                testResponseP(
+                    service,
+                    .content =
+                        "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
+                        "<ListBucketResult xmlns=\"http://s3.amazonaws.com/doc/2006-03-01/\">"
+                        "    <IsTruncated>false</IsTruncated>"
+                        "    <Version>"
+                        "        <Key>path/3/test_file</Key>"
+                        "        <LastModified>2024-08-04T02:54:10.000Z</LastModified>"
+                        "    </Version>"
+                        "    <Version>"
+                        "        <Key>path/3/test_file</Key>"
+                        "        <LastModified>2024-08-04T02:54:09.000Z</LastModified>"
+                        "        <Size>6</Size>"
+                        "        <VersionId>bbbb</VersionId>"
+                        "    </Version>"
+                        "   <CommonPrefixes>"
+                        "       <Prefix>path/3/test_path/</Prefix>"
+                        "   </CommonPrefixes>"
+                        "</ListBucketResult>");
+
+                testRequestP(service, s3, HTTP_VERB_GET, "/path/3/test_file?versionId=bbbb", .requesterPays = true);
+                testResponseP(service, .content = "123456");
+
+                TEST_RESULT_STR_Z(
+                    strNewBuf(storageGetP(storageNewReadP(s3, STRDEF("/path/3/test_file")))), "123456", "get file");
+
+                // -----------------------------------------------------------------------------------------------------------------
+                TEST_TITLE("get missing file with time limit");
+
+                testRequestP(service, s3, HTTP_VERB_GET, "/?delimiter=%2F&versions=", .requesterPays = true);
+                testResponseP(
+                    service,
+                    .content =
+                        "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
+                        "<ListBucketResult xmlns=\"http://s3.amazonaws.com/doc/2006-03-01/\">"
+                        "    <IsTruncated>false</IsTruncated>"
+                        "    <Version>"
+                        "        <Key>missing_file</Key>"
+                        "        <LastModified>2024-08-04T02:54:10.000Z</LastModified>"
+                        "    </Version>"
+                        "    <DeleteMarker>"
+                        "        <Key>missing_file</Key>"
+                        "        <LastModified>2024-08-04T02:54:09.000Z</LastModified>"
+                        "    </DeleteMarker>"
+                        "    <CommonPrefixes>"
+                        "       <Prefix>test_path/</Prefix>"
+                        "    </CommonPrefixes>"
+                        "</ListBucketResult>");
+
+                TEST_RESULT_PTR(
+                    storageGetP(storageNewReadP(s3, STRDEF("missing_file"), .ignoreMissing = true)), NULL, "missing file");
+
                 // -----------------------------------------------------------------------------------------------------------------
                 hrnServerScriptEnd(service);
             }
